/*
Software to interface an RPi to an SSD1306 OLED display. Apologies for
the cobbled-together hacky code. It'll need libgd to compile.
(C) 2012 Jeroen Domburg (jeroen AT spritesmods.com)

This program is free software: you can redistribute it and/or modify
t under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

/*** RPi GPIO access stuff ***/
//Copy-pasted from http://elinux.org/RPi_Low-level_peripherals

// Access from ARM Running Linux

#define BCM2708_PERI_BASE        0x20000000
#define GPIO_BASE                (BCM2708_PERI_BASE + 0x200000) /* GPIO controller */


#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <dirent.h>
#include <fcntl.h>
#include <assert.h>
#include <sys/mman.h>
#include <sys/types.h>
#include <sys/stat.h>

#include <unistd.h>

#define PAGE_SIZE (4*1024)
#define BLOCK_SIZE (4*1024)

int  mem_fd;
char *gpio_mem, *gpio_map;
char *spi0_mem, *spi0_map;


// I/O access
volatile unsigned *gpio;


// GPIO setup macros. Always use INP_GPIO(x) before using OUT_GPIO(x) or SET_GPIO_ALT(x,y)
#define INP_GPIO(g) *(gpio+((g)/10)) &= ~(7<<(((g)%10)*3))
#define OUT_GPIO(g) *(gpio+((g)/10)) |=  (1<<(((g)%10)*3))
#define SET_GPIO_ALT(g,a) *(gpio+(((g)/10))) |= (((a)<=3?(a)+4:(a)==4?3:2)<<(((g)%10)*3))

#define GPIO_SET *(gpio+7)  // sets   bits which are 1 ignores bits which are 0
#define GPIO_CLR *(gpio+10) // clears bits which are 1 ignores bits which are 0



//
// Set up a memory regions to access GPIO
//
void setup_io()
{
   /* open /dev/mem */
   if ((mem_fd = open("/dev/mem", O_RDWR|O_SYNC) ) < 0) {
      printf("can't open /dev/mem \n");
      exit (-1);
   }

   /* mmap GPIO */

   // Allocate MAP block
   if ((gpio_mem = malloc(BLOCK_SIZE + (PAGE_SIZE-1))) == NULL) {
      printf("allocation error \n");
      exit (-1);
   }

   // Make sure pointer is on 4K boundary
   if ((unsigned long)gpio_mem % PAGE_SIZE)
     gpio_mem += PAGE_SIZE - ((unsigned long)gpio_mem % PAGE_SIZE);

   // Now map it
   gpio_map = (unsigned char *)mmap(
      (caddr_t)gpio_mem,
      BLOCK_SIZE,
      PROT_READ|PROT_WRITE,
      MAP_SHARED|MAP_FIXED,
      mem_fd,
      GPIO_BASE
   );

   if ((long)gpio_map < 0) {
      printf("mmap error %d\n", (int)gpio_map);
      exit (-1);
   }

   // Always use volatile pointer!
   gpio = (volatile unsigned *)gpio_map;


} // setup_io





#include <stdio.h>
#include <gd.h>

#define SSD1306_SETCONTRAST 0x81
#define SSD1306_DISPLAYALLON_RESUME 0xA4
#define SSD1306_DISPLAYALLON 0xA5
#define SSD1306_NORMALDISPLAY 0xA6
#define SSD1306_INVERTDISPLAY 0xA7
#define SSD1306_DISPLAYOFF 0xAE
#define SSD1306_DISPLAYON 0xAF
#define SSD1306_SETDISPLAYOFFSET 0xD3
#define SSD1306_SETCOMPINS 0xDA
#define SSD1306_SETVCOMDETECT 0xDB
#define SSD1306_SETDISPLAYCLOCKDIV 0xD5
#define SSD1306_SETPRECHARGE 0xD9
#define SSD1306_SETMULTIPLEX 0xA8
#define SSD1306_SETLOWCOLUMN 0x00
#define SSD1306_SETHIGHCOLUMN 0x10
#define SSD1306_SETSTARTLINE 0x40
#define SSD1306_MEMORYMODE 0x20
#define SSD1306_COMSCANINC 0xC0
#define SSD1306_COMSCANDEC 0xC8
#define SSD1306_SEGREMAP 0xA0
#define SSD1306_CHARGEPUMP 0x8D
#define SSD1306_EXTERNALVCC 0x1
#define SSD1306_SWITCHCAPVCC 0x2


//Masks for the pin signals
#define LCD_CS 29
#define LCD_A0 27
#define LCD_SCK 31
#define LCD_SDI 28

//H & W of lcd. We use it in landscape mode here.
#define LCD_H 32
#define LCD_W 128

void setpin(int pin, int val) {
	if (val) GPIO_SET=(1<<pin); else GPIO_CLR=(1<<pin);
}

void claimpin(int pin) {
	INP_GPIO(pin);
	OUT_GPIO(pin);
}

//Raise CS
void spiDone(void) {
	setpin(LCD_CS, 1);
}

//Send a SPI byte
int spiSend(unsigned char byte, int isData) {
	unsigned char b=0;
	int x;
	setpin(LCD_CS, 0);
	setpin(LCD_A0, isData);
	for (x=0; x<8; x++) {
		if (byte&0x80) setpin(LCD_SDI, 1); else setpin(LCD_SDI, 0);
		setpin(LCD_SCK, 0);
		setpin(LCD_SCK, 1);
		byte<<=1;
	}
}


void spiFlush() {
	//not used here
}


#define ssd1306_command(x) spiSend(x, 0)
//Init the LCD
void lcdInit() {
    ssd1306_command(SSD1306_DISPLAYOFF); // 0xAE
    ssd1306_command(SSD1306_SETDISPLAYCLOCKDIV); // 0xD5
    ssd1306_command(0x80); // the suggested ratio 0x80
    ssd1306_command(SSD1306_SETMULTIPLEX); // 0xA8
    ssd1306_command(0x1F);
    ssd1306_command(SSD1306_SETDISPLAYOFFSET); // 0xD3
    ssd1306_command(0x0); // no offset
    ssd1306_command(SSD1306_SETSTARTLINE | 0x0); // line #0
    ssd1306_command(SSD1306_CHARGEPUMP); // 0x8D
    ssd1306_command(0x14);
    ssd1306_command(0xAF); // 0xAE
    ssd1306_command(SSD1306_MEMORYMODE); // 0x20
    ssd1306_command(0x00); // 0x0 act like ks0108
	ssd1306_command(SSD1306_SEGREMAP | 0x1);
    ssd1306_command(SSD1306_COMSCANDEC);
    ssd1306_command(SSD1306_SETCOMPINS); // 0xDA
    ssd1306_command(0x02);
    ssd1306_command(SSD1306_SETCONTRAST); // 0x81
    ssd1306_command(0x8F);
    ssd1306_command(SSD1306_SETPRECHARGE); // 0xd9
    ssd1306_command(0xF1);
    ssd1306_command(SSD1306_SETVCOMDETECT); // 0xDB
    ssd1306_command(0x40);
    ssd1306_command(SSD1306_DISPLAYALLON_RESUME); // 0xA4
    ssd1306_command(SSD1306_NORMALDISPLAY); // 0xA6
//    ssd1306_command(SSD1306_INVERTDISPLAY); // 0xA6
	ssd1306_command(SSD1306_DISPLAYON);
	spiFlush();
}

//Show the png-file called filename on the LCD
void lcdShowImage(char* filename) {
	FILE *f;
	int x, y, z, p, b;
	unsigned int col;
	gdImagePtr img;
	f=fopen(filename,"r");
	if (f==NULL) {
		perror(filename);
		return;
	}
	img=gdImageCreateFromPng(f);
	fclose(f);
	if (img==NULL) {
		printf("Couldn't load image: %s\n", filename);
		return;
	}


	ssd1306_command(SSD1306_SETLOWCOLUMN | 0x0); // low col = 0
	ssd1306_command(SSD1306_SETHIGHCOLUMN | 0x0); // hi col = 0
	ssd1306_command(SSD1306_SETSTARTLINE | 0x0); // line #0
	for (y=0; y<LCD_H; y+=8) {
		for (x=0; x<LCD_W; x++) {
			for (z=0; z<8; z++) {
				p=gdImageGetPixel(img, x, y+z);
				b>>=1;
				if (gdImageGreen(img, p)>128) b|=0x80;
			}
			spiSend(b, 1);
		}
	}

	//seems necessary
	for (y=0; y<(LCD_H*LCD_W/8); y++) spiSend(0,1);

	spiDone();
	spiFlush();
}

//Main function
int main(int argc, char **argv) {
	int x;
	setup_io();
	claimpin(LCD_CS);
	claimpin(LCD_A0);
	claimpin(LCD_SCK);
	claimpin(LCD_SDI);
	//Init lcd
	lcdInit();
	//Send all pngs on the command line
	for (x=1; x<argc; x++) {
		lcdShowImage(argv[x]);
	}
}

